package gui;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Vector;

import scorekeeping.Question;
import scorekeeping.VoteReceiver;

/**
 * A timer object is used in UpopVote as a swing component which
 * is literally a timer (displays the time remaining at points throughout
 * the game) and controls the flow through the game.
 *
 * @author Michael Gebauer
 */
public class GameFlowModel implements Runnable, ActionListener {

  public static final long serialVersionUID = 20050112L;
  public static final int DEFAULT_QUESTION_TIME = 60;

  // These variables define the current state of the game.
  public enum State {idle, registration, round_1, round_1_winner, inter_round_text, round_2, round_2_winner};
  private State state = State.idle;
  private int questionNumber = 0;
  private boolean isPaused = true;
  private int remainingTime = 0;

  // These variables are needed to control aspects of the GUI
  private UpopVote upopVote;
  private QuestionPanel qp;
  private Vector<Question> round1Questions, round2Questions;
  private Scoreboard sb;
  private TimerPanel tp;
  private Thread timerThread;

  @SuppressWarnings("unchecked")
  public GameFlowModel(UpopVote upopVote, Vector<Question> round1Questions, Vector<Question> round2Questions) {
    this.upopVote = upopVote;
    this.qp = upopVote.getQuestionPanel();
    this.sb = upopVote.getScoreboard();
    this.tp = upopVote.getTimerPanel();
    this.round1Questions = round1Questions;
    this.round2Questions = round2Questions;

    tp.addButtonListener(this);

    timerThread = new Thread(this, "Timer");
    timerThread.start();
  }

  /**
   * Returns the number of this question.
   * This is used as the window title.
   */
  public String getGameStateString() {
    switch(state) {
      case idle :
        return "Welcome to the Groupthink Specification Exercise";
      case registration :
        return "Registration mode";
      case round_1 :
        int number = questionNumber+1;
        int size = round1Questions.size();
        return "Question "+number+" of "+size;
      case round_1_winner :
        return "Groupthink Specification Exercise";
      case inter_round_text :
        return "Groupthink Specification Exercise";
      case round_2 :
        number = questionNumber+1;
        size = round2Questions.size();
        return "Question "+number+" of "+size;
      case round_2_winner :
        return "Groupthink Specification Exercise";
      default :
        throw new RuntimeException("Unexpected state type: "+state);
    }
  }

  public State getGameState() {
    return state;
  }

  /**
   * This method implements the required run method from the
   * Runnable interface.
   */
  public void run() {
    try {
      Thread currentThread = Thread.currentThread();
      while (timerThread == currentThread) {
        Thread.sleep(1000);
        decrementTimer();
      }
    } catch (InterruptedException e1) {
      // If the thread is interrupted, just restart:
      run();
    }
  }

  public void actionPerformed(ActionEvent e) {
    String command = e.getActionCommand();
    if (tp.next.equals(command)) {
      next();
    } else if (tp.pause.equals(command)) {
      pause();
    } else if (tp.addTime.equals(command)) {
      addTime();
    } else if (tp.removeTime.equals(command)) {
      removeTime();
    }
  }

  public synchronized void decrementTimer() {
    if (remainingTime > 1
        && !isPaused) {
      remainingTime = --remainingTime;
      tp.setTime(remainingTime);
    }
    if (remainingTime == 1) {
      if (state == State.round_1 || state == State.round_2) {
        remainingTime = 0;
        endQuestion();
      }
    }
    // else time is expired so do nothing
  }

  private void startQuestion() {
    assert state == State.round_1 || state == State.round_2;
    Question question;
    if (state == State.round_1) {
      question = round1Questions.get(questionNumber);
    } else {
      question = round2Questions.get(questionNumber);
    }

    sb.displayAllTransmitterIDs();
    qp.setText(question.toString(false));
    remainingTime = DEFAULT_QUESTION_TIME;
    tp.setTime(remainingTime);
    VoteReceiver.setDestination(question);
  }

  private void endQuestion() {
    assert state == State.round_1 || state == State.round_2;

    // stop receiving votes in the vote receiver
    VoteReceiver.setDestination(null);
    remainingTime = 0;
    tp.setTime(0);

    Question question;
    if (state == State.round_1) {
      question = round1Questions.get(questionNumber);
    } else {
      question = round2Questions.get(questionNumber);
    }

    // display the answer
    String questionText = question.toString(true);
    qp.setText(questionText);

    // show vote summaries
    sb.displayScoreSummary(question);
  }

  // The next command (triggered by the "Advance" button) is used to
  // advance the state of the game, regardless of what the current state
  // may be.  If the game is paused, this command will unpause.  If a
  // question is in progress, it will be ended.  If no question is in
  // progress, the next question will be initiated.
  public void next() {
    // if the game is paused, then unpause it
    if (isPaused)
      pause();

    switch(state) {
    case idle :
      state = State.registration;
      break;
    case registration :
      state = State.round_1;
      questionNumber = 0;
      startQuestion();
      break;
    case round_1 :
      if (remainingTime>0) {
        // finish the current question
        endQuestion();
      } else {
        questionNumber += 1;
        // if there's another question in the round,
        // then start it, else move to the next state
        int totalQuestions = round1Questions.size();
        if (questionNumber<totalQuestions) {
          startQuestion();
        } else {
          state = State.round_1_winner;
          sb.clearIdSection();
          Team[] winners = sb.getWinner();
          String text = "Congratulations to: <ul>";
          for (Team team: winners) {
            text = text + "<li>" + team.toString();
          }
          qp.setText(text+"</ul>");
        }
      }
      break;
    case round_1_winner :
      // immediately switch the game into inter_round_text mode
      state = State.inter_round_text;
      sb.displayMostImproved();
      // also display some text
      qp.setText("Round 2 changes" +
                 "<br>" +
                 "<ul>" +
                 "<li>Some answers are marked as wrong" +
                 "<ul>" +
                 "<li>Violates requirements</li>" +
                 "<li>Impossible behavior</li>" +
                 "</ul></li>" +
                 "<li>All point values are doubled" +
                 "<br>" +
                 "<li>Two winning teams" +
                 "<ul>" +
                 "<li>Highest overall score</li>" +
                 "<li>Most improved</li>" +
                 "</ul></li>" +
                 "</ul>");
      break;
    case inter_round_text :
      state = State.round_2;
      questionNumber = 0;
      startQuestion();
      break;
    case round_2 :
      if (remainingTime>0) {
        // finish the current question
        endQuestion();
      } else {
        questionNumber += 1;
        // if there's another question in the round,
        // then start it, else move to the next state
        int totalQuestions = round2Questions.size();
        if (questionNumber<totalQuestions) {
          startQuestion();
        } else {
          state = State.round_2_winner;
          sb.clearIdSection();
          Team[] winners = sb.getWinner();
          Team[] mostImproved = sb.getMostImproved();
          String text = "Congratulations to <ul>";
          for (Team team : winners) {
            text = text + "<li>" + team.toString() + " (high score)</li>";
          }
          for (Team team : mostImproved) {
            text = text + "<li>" + team.toString() + " (most improved)</li>";
          }
          qp.setText(text + "</ul>");
        }
      }
      break;
    case round_2_winner :
      break;
    default :
      throw new RuntimeException("Unexpected state type: "+state);
    }
    // for all cases, update the game state displayed on the title bar
    upopVote.setTitle(getGameStateString());
  }

  public void pause() {
    isPaused = !isPaused;
    VoteReceiver.setPaused(isPaused);
  }

  public void addTime() {
    if (remainingTime > 0 && remainingTime < 90) {
      remainingTime += 10;
      tp.setTime(remainingTime);
    }
  }

  public void removeTime() {
    if (remainingTime > 0) {
      if (remainingTime>10) {
        remainingTime -= 10;
        tp.setTime(remainingTime);
      } else {
        next();
      }
    }
  }
}

//  LocalWords:  UpopVote unpause li ul br
