package gui;

import static gui.UpopVote.MAX_TEAMS;
import static gui.UpopVote.MAX_TEAM_SIZE;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Vector;

import javax.swing.table.AbstractTableModel;

import scorekeeping.Question;
import scorekeeping.Vote;

/**
 * An implementation of a table model specifically for the the Scoreboard.
 *
 * @author Michael Gebauer
 *
 */
public class ScoreboardModel extends AbstractTableModel {

  static final long serialVersionUID = 20050112L;

  private boolean showMostImproved = false;

  // The data model:
  // The first column which holds team names.
  private Team[] teamColumn = Team.values();
  // The part of the scoreboard which holds player IDs, among other things
  private String[][] idSection = new String[MAX_TEAMS][MAX_TEAM_SIZE];
  // The total score columns
  private HashMap<Team,Double> scoreColumns = new HashMap<Team,Double>();
  // The most improved score columns
  private HashMap<Team,Double> improvedColumns = new HashMap<Team,Double>();
  // First half score for each team, but isn't displayed in the table
  private double[] firstHalfScore = new double[MAX_TEAMS];
  // Second half score for each team, but isn't displayed in the table
  private double[] secondHalfScore = new double[MAX_TEAMS];
  // One entry per participant.  True iff the participant is present
  // and participating in the game.
  private boolean[][] present = new boolean[MAX_TEAMS][MAX_TEAM_SIZE];

  public ScoreboardModel() {
    for (Team t : Team.values()) {
      scoreColumns.put(t,0.);
      improvedColumns.put(t,0.);
    }
  }

  public int getRowCount() {
    return MAX_TEAMS;
  }

  public int getColumnCount() {
    if (!showMostImproved)
      return idSection[0].length+3;
    else
      return idSection[0].length+5;
  }

  /**
   * Returns the value intended to appear in the location of the
   * JTable indicated by the arguments.
   *
   */
  @SuppressWarnings("unchecked")
  public Object getValueAt(int row, int column) {
    assert 0 <= row && row < getRowCount();
    assert 0 <= column && column < getColumnCount();

    if (column == 0)
      return teamColumn[row];
    else if (column <= MAX_TEAM_SIZE)
      return idSection[row][column-1];

    // if the column is in one of the scoring columns
    int teamColumn = MAX_TEAM_SIZE+1;
    int scoreColumn = MAX_TEAM_SIZE+2;
    if (column == teamColumn || column == scoreColumn) {
      // get the element of index row from the hashmap by converting to an ArrayList and then sorting
      ArrayList<Map.Entry<Team,Double>> scores = new ArrayList<Map.Entry<Team,Double>>(scoreColumns.entrySet());
      Collections.sort(scores, new ScoreboardComparator());
      Map.Entry<Team,Double> entry = scores.get(row);
      if (column == teamColumn)
        return entry.getKey();
      else return Math.round(entry.getValue());
    }

    // if the column is one of the most improved columns
    teamColumn = MAX_TEAM_SIZE+3;
    scoreColumn = MAX_TEAM_SIZE+4;
    if (column == teamColumn || column == scoreColumn) {
      // get the element of index row from the hashmap by converting to an ArrayList and then sorting
      ArrayList<Map.Entry<Team,Double>> scores = new ArrayList<Map.Entry<Team,Double>>(improvedColumns.entrySet());
      Collections.sort(scores, new ScoreboardComparator());
      Map.Entry<Team,Double> entry = scores.get(row);
      if (column == teamColumn)
        return entry.getKey();
      // return the value rounded to 2 decimal places
      else return Math.rint(entry.getValue()*100)/100;
    }

    else throw new RuntimeException("Tried to get row "+row+", col "+column+", but doesn't exist.");
  }

  // This is used only for setting the column names in the model.
  // Because JTable.getColumn takes a column name as an argument, all
  // column names must be unique -- this is why the two blank ones have
  // different numbers of spaces.
  public String getColumnName(int columnNumber) {
    assert columnNumber<getColumnCount();
    if (columnNumber == 0)
      return "Team";
    else if (columnNumber <= MAX_TEAM_SIZE)
      return String.valueOf(columnNumber);
    else if (columnNumber == (MAX_TEAM_SIZE+1))
      return "Score";
    else if (columnNumber == (MAX_TEAM_SIZE+2))
      return "";
    else if (columnNumber == (MAX_TEAM_SIZE+3) && showMostImproved)
      return "Improvement";
    else if (columnNumber == (MAX_TEAM_SIZE+4) && showMostImproved)
      return " ";
    else
      throw new RuntimeException("Can't get column name for " +
          "column "+columnNumber+".  There are only " +
          getColumnCount()+" columns.");

  }

  // Returns an array of the team(s) with the highest total score.
  // The array will contain all teams tied for first places.
  public Team[] getWinner() {
    int teamCol = MAX_TEAM_SIZE+1;
    int scoreCol = MAX_TEAM_SIZE+2;
    int winningScore = ((Long) getValueAt(0,scoreCol)).intValue();
    // Return all teams that have the winning score
    Vector<Team> winners = new Vector<Team>();
    for (int row=0; row<scoreColumns.size(); row++) {
      Long cellValue = (Long) getValueAt(row,scoreCol);
      int thisScore = cellValue.intValue();
      if (thisScore == winningScore) {
        Team team = (Team) getValueAt(row,teamCol);
        winners.add(team);
      }
    }
    return winners.toArray(new Team[winners.size()]);
  }

  // Returns an array of the team(s) with the most improvement.
  // The array will contain all teams tied for first places.
  public Team[] getMostImproved() {
    assert showMostImproved;
    int teamCol = MAX_TEAM_SIZE+3;
    int scoreCol = MAX_TEAM_SIZE+4;
    double winningRatio = ((Double) getValueAt(0,scoreCol)).doubleValue();
    // Return all teams that have the winning ratio.
    Vector<Team> winners = new Vector<Team>();
    for (int row=0; row<scoreColumns.size(); row++) {
      Double cellValue = (Double) getValueAt(row,scoreCol);
      double thisScore = cellValue.doubleValue();
      if (thisScore == winningRatio) {
        Team team = (Team) getValueAt(row,teamCol);
        winners.add(team);
      }
    }
    return winners.toArray(new Team[winners.size()]);
  }

  // Displays all transmitter ids for players who are present.
  public synchronized void displayAllTransmitterIDs() {
    // for each team and each player on that team
    for (int teamno=0; teamno<MAX_TEAMS; teamno+=1)
      for (int playerno=0; playerno<MAX_TEAM_SIZE; playerno+=1) {
        // lookup the id
        int playerID = Team.getTeam(teamno).getIDs()[playerno];
        // ensure that the player is participating in this game
        if (present[teamno][playerno])
          idSection[teamno][playerno]=String.valueOf(playerID).substring(4);
        else
          idSection[teamno][playerno]="";
      }
    fireTableRowsUpdated(0,idSection.length);
  }

  // clear the id section
  public void clearIdSection() {
    for (int teamno=0; teamno<idSection.length; teamno+=1) {
      for (int playerno=0; playerno<idSection[teamno].length; playerno+=1) {
        idSection[teamno][playerno] = "";
      }
    }
    fireTableRowsUpdated(0,idSection.length);
  }

  // Players and teams are numbered starting at 0.
  public void display(String s, int team, int player) {
    assert 0 <= team && team < MAX_TEAMS : "team " + team;
    assert 0 <= player && player < MAX_TEAM_SIZE : "player " + player;

    idSection[team][player] = s;
    fireTableCellUpdated(team,player+1);
  }

  // Clears one transmitter ID.  Does not clear
  // if the transmitter ID is not being displayed.
  //
  // *Note, the team is passed as an argument simply so that
  // it doesn't get looked up more than it needs to be.  This
  // was an efficiency decision, but I'm not sure how much it
  // helps.  The team argument might be removed without any
  // performance loss.
  public synchronized void removeID(Team t, int id) {
    // remove the id
    int teamno = t.getNumber();
    int playerno = t.getPlayerNumber(id);
    display("", teamno, playerno);
  }

  // Displays the score for a specified team.  This method must be called
  // at least once per team, per question.  The score value passed
  // as a parameter will be the value added.  Specifically, if the
  // score is to be doubled, the parameter should be doubled.
  public void displayScoreUpdate(Team t, double score) {
    int teamno = t.getNumber();
    for (int playerno=1; playerno<MAX_TEAM_SIZE; playerno++) {
      idSection[teamno][playerno] = "";
    }

    // update the score columns
    double newScore = score + scoreColumns.get(t);
    scoreColumns.put(t,newScore);

    // update the most improved columns
    if (showMostImproved && firstHalfScore[teamno]!=0) {
      secondHalfScore[teamno] += score;
      improvedColumns.put(t,secondHalfScore[teamno] / firstHalfScore[teamno]);
    }

    // display this round's score in column #1
    if (!idSection[teamno][0].contains("pts")) {
      String scoreString = String.valueOf(Math.round(score))+" pts";
      idSection[teamno][0] = scoreString;
    }

    fireTableDataChanged();
  }

  // This method gets called between the first and second halves
  // of the game.
  public void displayMostImproved() {
    showMostImproved = true;
    fireTableStructureChanged();

    // store the scores from the first half
    int teamColumn = MAX_TEAM_SIZE+1;
    int scoreColumn = MAX_TEAM_SIZE+2;
    for (int row=0; row<MAX_TEAMS; row++) {
      Team team = (Team) getValueAt(row,teamColumn);
      int teamno = team.getNumber();
      long score =(Long) getValueAt(row,scoreColumn);
      firstHalfScore[teamno] = score;
    }

    clearIdSection();
  }

  public void displayScoreSummary(Question question) {
    // begin by clearing all the ids
    for (Team t: Team.values())
      for (int id : t.getIDs())
        if (t.getPlayerNumber(id)!=0)
          removeID(t,id);

    Vote[][] votes = question.getVotes();
    for (Team t: Team.values()) {
      int teamno = t.getNumber();
      // Build the vote histogram (this vector will have one
      // entry for each possible vote (0-9) and will store
      // the number of votes received.
      int[] countArray = new int[10];
      for (Vote vote : votes[teamno]) {
        if (vote != null) {
          int value = vote.getValue();
          countArray[value] += 1;
        }
      }
      // Sort the vote histogram by count.
      //
      // Note that i=2 begins the loop.  This is because
      // the first two columns of idSection are used
      // to display the score.
      for (int i=2; i<MAX_TEAM_SIZE; i++) {
        int max = -1;
        int index = -1;
        for (int j=0; j<MAX_TEAM_SIZE; j++) {
          if (countArray[j]>max) {
            max = countArray[j];
            index = j;
          }
        }

        // display number of agreeing votes
        if (index>-1 && max>0) {
          countArray[index] = -1;
          if (question.isWrongAnswer(index-1)) {
            idSection[teamno][i] = max+": "+index+"s*";
          } else if (question.isStrangeAnswer(index-1)) {
            idSection[teamno][i] = max+": "+index+"s?";
          } else {
            idSection[teamno][i] = max+": "+index+"s";
          }
        }
      }
    }
  }

  // *Note, the team is passed as an argument simply so that
  // it doesn't get looked up more than it needs to be.  This
  // was an efficiency decision, but I'm not sure how much it
  // helps.  The team argument might be removed without any
  // performance loss.
  public void registerPlayer(Team t, Vote v) {
    int teamno = t.getNumber();
    int playerID = t.getPlayerNumber(v.getId());
    String voteString = String.valueOf(v.getId()).substring(3) + " : " + v.getValue();
    present[teamno][playerID] = true;
    display(voteString,teamno,playerID);

    UpopVote.print("ScoreboardModel: registered player: ","\t"+v,"\t"+t);
  }
}

//  LocalWords:  iff JTable hashmap col pts idSection ScoreboardModel
